@@ -19,18 +19,11 @@ require Exporter;
 our @ISA = qw(Exporter);
 our @EXPORT = qw(
 	     doPropertySetInfo
-	     emitContextNames
-	     emitContextProperties
 	     emitEvalResultAsProperty
 	     emitEvaluatedPropertyGetInfo
 	     figureEncoding
-	     getContextProperties
 	     getFullPropertyInfoByValue
-	     getPropertyInfo
-	     GlobalVars
-	     LocalVars
-	     FunctionArguments
-	     PunctuationVariables
+	     makeFullPropertyName
 	     );
 our @EXPORT_OK = ();
 
@@ -42,36 +35,15 @@ use DB::DbgrCommon;
 # Internal sub declarations
 
 sub adjustLongName($$$);
-sub makeFullName($$);
+sub makeFullPropertyName($$);
 
 # And recursively-called exported routines:
 
 sub figureEncoding($);
 sub getFullPropertyInfoByValue($$$$$$);
 
-# Constants
-
-use constant LocalVars => 0;
-use constant GlobalVars => 1;
-use constant FunctionArguments => 2;
-use constant PunctuationVariables => 3;
-
 our $ldebug = 0;
 
-# Private Data
-
-my %contextProperties = (
-    Globals => GlobalVars,
-    Locals => LocalVars,
-    Arguments => FunctionArguments,
-    Special => PunctuationVariables,
-);
-my @contextProperties = sort { $contextProperties{$a} <=> $contextProperties{$b} } keys %contextProperties;
-
-my @_punctuationVariables = ('$_', '$?', '$@', '$.', '@+', '@-', '$+', '$!', '$$', '$0');
-# $`, $& and $' are special-cased to avoid the performance penalty
-my @_rxPunctuationVariables = ('`', '&', '\'');
-
 # Exported subs
 
 =head1 postconditions
@@ -107,65 +79,6 @@ sub doPropertySetInfo($$$) {
     }
 }
 
-sub emitContextNames($$) {
-    my ($cmd, $transactionID) = @_;
-    my $res = sprintf(qq(%s\n<response %s command="%s" 
-			 transaction_id="%s" >),
-		      xmlHeader(),
-		      namespaceAttr(),
-		      $cmd,
-		      $transactionID);
-    # todo: the spec suggests that locals be the default,
-    # but globals (package globals) make more sense for Perl.
-    for (my $i = 0; $i <= $#contextProperties; $i++) {
-	$res .= sprintf(qq(<context name="%s" id="%d" />\n),
-			$contextProperties[$i],
-			$i);
-    }
-    $res .= "\n</response>\n";
-    printWithLength($res);
-}
-
-sub emitContextProperties($$$$;$) {
-    my ($cmd,
-	$transactionID,
-	$context_id,
-	$nameValuesARef,
-	$maxDataSize) = @_;
-    
-    my $res = sprintf(qq(%s\n<response %s command="%s"
-			 context_id="%d"
-			 transaction_id="%s" >),
-		      xmlHeader(),
-		      namespaceAttr(),
-		      $cmd,
-		      $context_id,
-		      $transactionID);
-    my @results = @$nameValuesARef;
-    my $numVars = scalar @results;
-    for (my $i = 0; $i < $numVars; $i++) {
-	my $result = $results[$i];
-	my $name = $result->[0];
-	my $val = $result->[1];
-	eval {
-	    my $property = getFullPropertyInfoByValue($name,
-						      $name,
-						      $val,
-						      $maxDataSize,
-						      0,
-						      0);
-	    # dblog("emitContextProperties: getFullPropertyInfoByValue => $property") if $ldebug;
-	    $res .= $property;
-	};
-	if ($@) {
-	    dblog("emitContextProperties: error [$@]") if $ldebug;
-	}
-    }
-    $res .= "\n</response>";
-    printWithLength($res);
-}
-
-
 sub emitEvaluatedPropertyGetInfo($$$$$$$) {
     my ($cmd,
 	$transactionID,
@@ -181,12 +94,12 @@ sub emitEvaluatedPropertyGetInfo($$$$$$$) {
 		      namespaceAttr(),
 		      $cmd,
 		      $transactionID);
+    my $finalName = $nameAndValue->[NV_NAME];
     my $finalVal = $nameAndValue->[NV_VALUE];
 
     $res .= '>' if $cmd ne 'property_value';
     my $startTag = $cmd ne 'property_value' ? '<property' : '';
     my $endTag = $cmd ne 'property_value' ? '</property>' : '';
-    my $finalName = makeFullName($property_long_name, $propertyKey);
     $res .= _getFullPropertyInfoByValue($startTag, $endTag,
 					$propertyKey || $finalName, # name
 					$finalName,
@@ -199,66 +112,6 @@ sub emitEvaluatedPropertyGetInfo($$$$$$$) {
     printWithLength($res);
 }
 
-# Return a ref to an array of [name, value, needValue] triples
-#
-# Some values can be evaluated in this scope, but non-package values
-# and locals at the top-level will need to be evaluated in the
-# debugger's main loop.
-
-sub getContextProperties($$) {
-    my ($context_id, $packageName) = @_;
-
-    # Here just show the top-level.
-    local $settings{max_depth}[0] = 0;
-    if ($context_id == GlobalVars) {
-	require B if $] >= 5.010;
-	# Globals
-	# Variables on the calling frame
-	my $stash = \%{"${packageName}::"};
-	my @results;
-	for my $key (keys %$stash) {
-	    next if $key =~ /^(?:_<|[^0a-zA-Z_])/;
-	    next if $key =~ /::$/;
-	    my $glob = \($stash->{$key});
-	    my ($has_scalar, $is_glob);
-	    if ($] >= 5.010) {
-		my $gv = B::svref_2object($glob);
-		$is_glob = $gv->isa("B::GV");
-		$has_scalar = $is_glob && !$gv->SV->isa('B::SPECIAL');
-	    } else {
-		$is_glob = 1;
-		$has_scalar = defined ${*{$glob}{SCALAR}};
-	    }
-	    my $array = $is_glob && *{$glob}{ARRAY};
-	    my $hash = $is_glob && *{$glob}{HASH};
-	    next unless $has_scalar || $array || $hash;
-	    push @results, ["\$${key}", ${*{$glob}{SCALAR}}, 0] if $has_scalar;
-	    push @results, ["\@${key}", $array, 0] if $array;
-	    push @results, ["\%${key}", $hash, 0] if $hash;
-	}
-	return \@results;
-    } elsif ($context_id == PunctuationVariables) {
-	my ($packageName, $filename, $line) = caller(1);
-	my %vals;
-	my @results;
-	foreach my $pv (@_punctuationVariables) {
-	    push (@results, [$pv, undef, 1]);
-	}
-	foreach my $pv (@_rxPunctuationVariables) {
-	    # somebody might use @' and trigger the condition, I'm willing to risk it
-	    push (@results, ["\$$pv", undef, exists $main::{$pv} ? 1 : 0]);
-	}
-	return \@results;
-	
-    } else {
-	die sprintf("code:%d:error:%s",
-		    302,
-		    ("Not ready to evaluate "
-		     . $contextProperties[$context_id]
-		     . ' variables'));
-    }
-}
- 
 sub _truncateIfNecessary {
     my($res, $maxDataSize, $stripOuterBrackets) = @_;
     if ($stripOuterBrackets && $res =~ /^([\[\{\(\<]).*([\]\}\)\>])$/) {
@@ -281,12 +134,13 @@ sub _truncateIfNecessary {
     return $res;
 }
 
-sub emitEvalResultAsProperty($$$$$) {
+sub emitEvalResultAsProperty($$$$$$) {
     my ($cmd,
 	$transactionID,
 	$property_long_name,
 	$valRefs,
-	$maxDataSize) = @_;
+	$maxDataSize,
+	$pageIndex) = @_;
     my $res = sprintf(qq(%s\n<response %s command="%s" 
 			 transaction_id="%s" >),
 		      xmlHeader(),
@@ -297,22 +151,13 @@ sub emitEvalResultAsProperty($$$$$) {
 				       $property_long_name,
 				       $valRefs,
 				       $maxDataSize,
-				       0, # page
+				       $pageIndex,
 				       0, # current depth
 				       );
     $res .= "\n</response>";
     printWithLength($res);
 }
 
-sub getPropertyInfo($$) {
-    my ($property_long_name, $propertyKey) = @_;
-
-    # Invariant: FunctionArguments are handled by the caller.
-    
-    my $finalName = makeFullName($property_long_name, $propertyKey);
-    return [$finalName, undef, 1];
-}
-   
 sub propertyTagSpacer($) {
     my ($currentDepth) = @_;
     return ("\n" . ('  ' x $currentDepth));
@@ -369,6 +214,8 @@ sub _getFullPropertyInfoByValue {
 	$pageIndex,
 	$currentDepth,
 	) = @_;
+    require Scalar::Util;
+
     # dblog("getFullPropertyInfoByValue: (@_)\n");
     my $encoding;
     my $res = $startTag;
@@ -389,11 +236,10 @@ sub _getFullPropertyInfoByValue {
     my $hasChildren = 0;
     my $numChildren;
     my $className;
-    my ($h1, $h2) = split(//, $fullname, 2);
     my $encVal = undef;
     my $encValLength = undef;
-    my $refstr = "";
     my $address;
+    my $hasValue = 1;
     my $variableGroup = -1;
     use constant VARIABLE_GROUP_ARRAY => 1;
     use constant VARIABLE_GROUP_HASH => 2;
@@ -402,52 +248,36 @@ sub _getFullPropertyInfoByValue {
     } else {
 	# Unlike getPropertyInfo, this is where we find
 	# arrays and hashes
-	if ($refstr = ref $val) {
-	    my $stringifiedVal = "" . $val;
-	    if ($refstr =~ /^ARRAY/) {
-		$typeString = $refstr;
-		$variableGroup = VARIABLE_GROUP_ARRAY;
-		$numChildren = scalar @$val;
-		$hasChildren = $numChildren >= 1;
-		($address) = ($stringifiedVal =~ m/^ARRAY\(0x(.*)\)$/i);
-	    } elsif ($refstr =~ /^HASH/) {
-		$typeString = $refstr;
+	if (my $reftype = Scalar::Util::reftype($val)) {
+	    my $refstr = ref $val;
+	    $className = Scalar::Util::blessed($val);
+	    $address = Scalar::Util::refaddr($val);
+	    $typeString = defined $className ? $className : $reftype;
+	    $hasValue = 0;
+	    if ($reftype eq 'HASH') {
 		$variableGroup = VARIABLE_GROUP_HASH;
-		$numChildren = scalar keys %$val;
-		$hasChildren = $numChildren >= 1;
-		($address) = ($stringifiedVal =~ m/^HASH\(0x(.*)\)$/i);
-	    } elsif ($refstr =~ /^Regexp/) {
+		$numChildren = keys %$val;
+	    } elsif ($reftype eq 'ARRAY') {
+		$variableGroup = VARIABLE_GROUP_ARRAY;
+		$numChildren = @$val;
+	    } elsif (defined $className || $reftype eq 'SCALAR') {
+		# object but not array/hash: it's a scalar
+		$numChildren = 1;
+	    }
+	    if ($reftype eq 'REGEXP' || (
+		    $] < 5.012 &&
+		    $reftype eq 'SCALAR' &&
+		    $className eq 'Regexp' &&
+		    _hasQrMagic($val))) {
 		# Special-case -- only one in Perl?
-		$typeString = 'Regexp';
-		$numChildren = 0;
-		$hasChildren = 0;
 		$val = substr("$val", 0, $maxDataSize);
 		($encoding, $encVal) = figureEncoding($val);
 		$res .= sprintf(qq( encoding="%s"), $encoding);
 		$encValLength = length($encVal);
-		$refstr = undef;
-	    } else {
-		my $overloadedRefStr;
-		($typeString, $numChildren, $className, $overloadedRefStr) = analyzeVal($val);
-		if ($overloadedRefStr) {
-		    $refstr = $overloadedRefStr;
-		}
-		$hasChildren = $numChildren && 1;
-		if ($className) {
-		    $typeString = $className;
-		    if ($maxDataSize >= 0 && length($val) > $maxDataSize) {
-			# dblog("getFullPropertyInfoByValue: truncating data down to $maxDataSize bytes:\n[$val]\n");
-			$val = substr($val, 0, $maxDataSize);
-		    }
-		    ($encoding, $encVal) = figureEncoding($val);
-		    $res .= sprintf(qq( encoding="%s"), $encoding);
-		    $encValLength = length($encVal);
-		} else {
-		    ($address) = ($stringifiedVal =~ m/^.+\(0x(.*)\)$/i);
-		}
+		$numChildren = 0;
+		$hasValue = 1;
 	    }
 	} else {
-	    $hasChildren = 0;
 	    # It's a scalar -- get the underlying value and classify.
 	    # First convert wide chars to utf-8
 	    my $val2 = nonXmlChar_Encode($val);
@@ -457,6 +287,7 @@ sub _getFullPropertyInfoByValue {
 	    $encValLength = length($encVal);
 	    $typeString = getCommonType($val2);
 	}
+	$hasChildren = !!$numChildren;
     }
     $res .= sprintf(qq( type="%s"), xmlAttrEncode($typeString));
     $res .= qq( constant="0");
@@ -478,9 +309,7 @@ sub _getFullPropertyInfoByValue {
 	    my $childrenPerPage = $settings{max_children}[0];
 	    my $startIndex = $pageIndex * $childrenPerPage;
 	    my $endIndex = $startIndex + $childrenPerPage - 1;
-	    if ($variableGroup == VARIABLE_GROUP_ARRAY
-		|| $refstr =~ /=ARRAY\(0x.*\)/
-		|| "$val" =~ /=ARRAY\(0x.*\)/) {
+	    if ($variableGroup == VARIABLE_GROUP_ARRAY) {
 		my $arraySize = scalar @$val;
 		#### ???? $res .= sprintf(qq( numchildren="%d"), $arraySize);
 		$res .= qq(>);
@@ -504,11 +333,8 @@ sub _getFullPropertyInfoByValue {
 			$res .= "$innerProp";
 		    }
 		}
-	    } elsif ($variableGroup == VARIABLE_GROUP_HASH
-		     || $refstr =~ /=HASH\(0x.*\)/
-		     || "$val" =~ /=HASH\(0x.*\)/) {
-		my %hval = %$val;
-		my @keys = sort keys %hval;
+	    } elsif ($variableGroup == VARIABLE_GROUP_HASH) {
+		my @keys = sort keys %$val;
 		my $arraySize = scalar @keys;
 		#### ???? $res .= sprintf(qq( numchildren="%d"), $arraySize);
 		$res .= qq(>);
@@ -559,7 +385,7 @@ sub _getFullPropertyInfoByValue {
 	$res .= qq( size="$encValLength") if defined $encValLength;
 	$res .= qq( >);
 	$res .= _getFullPropertyInfoByValue_emitNames($maxDataSize, %b_attr);
-	if ($refstr || !defined($encVal)) {
+	if (!$hasValue || !defined($encVal)) {
 	    # Do nothing
 	} elsif ($DB::xdebug_no_value_tag) {
 	    $res .= $encVal;
@@ -575,6 +401,21 @@ sub _getFullPropertyInfoByValue {
     return $res;
 }
 
+sub _hasQrMagic {
+    my ($val) = @_;
+
+    require B;
+
+    my $b = B::svref_2object($val);
+
+    return unless $b->isa('B::PVMG');
+    for (my $m = $b->MAGIC; $m; $m = $m->MOREMAGIC) {
+        return 1 if $m->TYPE eq 'r';
+    }
+
+    return 0;
+}
+
 sub _getFullPropertyInfoByValue_emitNames {
     my ($maxDataSize, %b_attr) = @_;
     my $ret = "";
@@ -588,66 +429,6 @@ sub _getFullPropertyInfoByValue_emitNames {
     return $ret;
 }
 
-# Called only by wrapVars, so it does no expansion
-# It's only used to provide info on the variables for 
-# a given context
-
-sub analyzeVal($) {
-    my ($val) = @_;
-    if (!defined $val) {
-	return ('null', 0, undef);
-    } elsif (!(ref $val)) {
-	return ('null', 0, undef);
-    }
-    my $refstr = ref $val;
-    if ($refstr =~ /^ARRAY/) {
-	return ($refstr, scalar @$val, undef);
-    } elsif ($refstr =~ /^HASH/) {
-	return ($refstr, scalar keys %$val, undef);
-    } elsif ($refstr =~ /^REF/ || $refstr =~ /^SCALAR/) {
-	return ($refstr, 1, undef);
-    } elsif ($refstr =~ /^CODE/) {
-	return ('CODE', 0, undef);
-    } elsif ("$val" =~ /$refstr=/) {
-	my $strVal = "$val";
-        my $typeString = $refstr;
-        $typeString =~ s/\(0x\w+\)//;
-	if ($strVal =~ /=HASH\(0x\w+\)/) {
-	    return ($typeString, scalar keys %$val, $refstr);
-	} elsif ($strVal =~ /=ARRAY\(0x\w+\)/) {
-	    return ($typeString, scalar @$val, $refstr);
-	} else {
-	    return ($typeString, 0, $refstr);
-	}
-    } elsif (overload::Overloaded($val)) {
-	my $overloadedRefStr;
-	if ($overloadedRefStr = overload::StrVal($val)) {
-	    dblog('qqq:analyzeVal ' . __LINE__) if $ldebug;
-	    dblog("analyzeVal -- overloaded [$val] => $overloadedRefStr") if $ldebug;
-	    my $className;
-	    if ($overloadedRefStr =~ /^(.*)=/) {
-		$className = $1;
-		dblog("  \$className = $className, \$overloadedRefStr = $overloadedRefStr") if $ldebug;
-		if ($overloadedRefStr =~ /=HASH\(0x\w+\)/) {
-		    return ('HASH', scalar keys %$val, $className, $overloadedRefStr);
-		} elsif ($overloadedRefStr =~ /=ARRAY\(0x\w+\)/) {
-		    return ('ARRAY', scalar @$val, $className, $overloadedRefStr);
-		} else {
-		    return ('SCALAR', 0, $className, $overloadedRefStr);
-		}
-	    }
-	    dblog("Could pull className out of refstr [$overloadedRefStr]") if $ldebug;
-	}
-	# It's some other kind of bizarre overloaded operator.
-	dblog('qqq:analyzeVal ' . __LINE__) if $ldebug;
-	return ('SCALAR', 0, $overloadedRefStr);
-    } else {
-	# Return whatever it is.
-	$refstr =~ s/=.*//;
-	return ($refstr, 1, undef);
-    }
-}
-
 #############################################################################
 
 # Internal subs
@@ -693,14 +474,16 @@ sub adjustLongName($$$) {
     }
 }
 
-sub makeFullName($$) {
+sub makeFullPropertyName($$) {
     my ($property_long_name, $propertyKey) = @_;
     if (!$propertyKey) {
-	return $property_long_name;
+	return ($property_long_name, undef);
     } elsif ($property_long_name =~ /^[\@\%](.*)/) {
-	return sprintf(q($%s%s), $1, $propertyKey);
+	return (sprintf(q($%s%s), $1, $propertyKey), $propertyKey);
+    } elsif ($propertyKey =~ /^->/) {
+	return (sprintf(q(%s%s), $property_long_name, $propertyKey), $propertyKey);
     } else {
-	return sprintf(q(%s->%s), $property_long_name, $propertyKey);
+	return (sprintf(q(%s->%s), $property_long_name, $propertyKey), "->$propertyKey");
     }
 }
 
